       program jco2jco

! --   Program jco2jco reads a JCO file corresponding to one PEST control file
!      and rewrites it for another.

       implicit none
       integer j,k,nespar1,itemp1,itemp2,bn,ies,irow,jes,ipp,
     +         n,ierr,i,npar1,nobs1,npargp1,nprior1,icount,n1,
     +         npar2,nobs2,npargp2,nprior2,nespar2,iobs,kes,
     +         ifail,nobs3,ii,jj,idifflag,jjj
       integer lw(2),rw(2)
       integer, allocatable :: itrans1(:),itrans2(:)
       double precision dtemp,dtemp1,logfac,ilogfac,ratio,ratio1,
     + ratio2,dd1,dd2,scalefac,rstemp1,rstemp2
       double precision, allocatable :: pval1(:),pval2(:),scale1(:),
     + scale2(:),offset1(:),offset2(:)
       double precision, allocatable :: x1(:,:),x2(:,:),x3(:,:)
       character*12 aversion
       character*12 aapar,aspar1,aspar2
       character*15 atemp,atemp1,atemp2,aext
       character*20 aaobs
       character*40 aprefix
       character*12, allocatable :: apar1(:),apar2(:)
       character*20, allocatable :: aobs1(:),aobs2(:)
       character*200 pestfile1,pestfile2,jcofile1,jcofile2,afile,bfile
       character*400 cline,errmsg

! -- Initialisation

       logfac=log(10.0d0)
       ilogfac=1.0d0/logfac
       idifflag=0

       include 'version.inc'
       write(6,10) trim(aversion)
10     format(' JCO2JCO Version ',a,'. Watermark Numerical Computing.')
#ifdef NO_CMDLINE
       write(*,*)
1      write(*,'(a)') ' ENTER COMMAND LINE ARGUMENTS > '
       read(*,'(a)',ERR=1) CLINE
#else
       call pgetcl(cline)
#endif

! -- The command line is parsed.

       if(cline.eq.' ') go to 9000
       call spacesub(cline)
       call linspl(ifail,3,lw,rw,cline)
       if(ifail.eq.0) go to 9000
       call linspl(ifail,2,lw,rw,cline)
       if(ifail.ne.0) go to 9000
       pestfile1=cline(lw(1):rw(1))
       pestfile2=cline(lw(2):rw(2))
       call remchar(pestfile1,char(211))
       call remchar(pestfile2,char(211))
       n=len_trim(pestfile1)
       n1=n-3
       if(n1.lt.1)n1=1
       aext=pestfile1(n1:n)
       call lowcase(aext)
       if(aext.eq.'.pst')pestfile1(n1:n)=' '
       jcofile1=pestfile1
       n=len_trim(pestfile1)
       pestfile1(n+1:)='.pst'
       jcofile1(n+1:)='.jco'
       n=len_trim(pestfile2)
       n1=n-3
       if(n1.lt.1)n1=1
       aext=pestfile2(n1:n)
       call lowcase(aext)
       if(aext.eq.'.pst')pestfile2(n1:n)=' '
       jcofile2=pestfile2
       n=len_trim(pestfile2)
       pestfile2(n+1:)='.pst'
       jcofile2(n+1:)='.jco'

! -- Input files are opened.

       open(unit=10,file=pestfile1,status='old',iostat=ierr)
       if(ierr.ne.0)then
         call addquote(pestfile1,afile)
         write(errmsg,110) trim(afile)
110      format('Cannot open PEST control file ',a,'.')
         call write_message(errmsg)
         stop
       end if
#ifdef LAHEY
#ifdef UNIX
       open(unit=15,file=jcofile1,form='unformatted',
     + status='old',iostat=ierr)
#else
       open(unit=15,file=jcofile1,form='unformatted',
     + access='transparent',status='old',iostat=ierr)
#endif
#else
       open(unit=15,file=jcofile1,form='unformatted',
     + status='old',iostat=ierr)
#endif
       if(ierr.ne.0)then
         call addquote(jcofile1,afile)
         write(errmsg,120) trim(afile)
120      format('Cannot open unformatted Jacobian file ',a,'.')
         call write_message(errmsg)
         stop
       end if

! -- Read the first control file - parameter and observation data only
!    (not prior information).

       do i=1,3
         read(10,*,err=9100,end=9100)
       end do
       read(10,*,err=9100,end=9100) npar1,nobs1,npargp1,nprior1
       allocate(apar1(npar1),pval1(npar1),itrans1(npar1),
     + aobs1(nobs1),scale1(npar1),offset1(npar1),stat=ierr)
       if(ierr.ne.0)then
         write(errmsg,125)
125      format('Cannot allocate sufficient memory to continue ',
     +   'JCO2JCO execution.')
         go to 9890
       end if
       do
         read(10,'(a)',err=9100,end=130) cline
         call lowcase(cline)
         if(index(cline,'* parameter data').ne.0) go to 150
       end do
130    continue
       call addquote(pestfile1,afile)
       write(errmsg,140) trim(afile)
140    format('Cannot find "parameter data" section of PEST ',
     + 'control file ',a,'.')
       go to 9890
150    do i=1,npar1
         read(10,*,err=9100,end=9100) apar1(i),atemp,atemp1,pval1(i),
     +   dtemp,dtemp,atemp1,scale1(i),offset1(i)
         call lowcase(apar1(i))
         call lowcase(atemp)
         if(atemp.eq.'log')then
           itrans1(i)=1
         else if(atemp.eq.'none')then
           itrans1(i)=0
         else if(atemp.eq.'fixed')then
           itrans1(i)=-100000
         else if(atemp.eq.'tied')then
           itrans1(i)=-1
         else
           go to 9100
         end if
       end do
       icount=0
       do i=1,npar1
         if(itrans1(i).eq.-1)icount=icount+1
       end do
       if(icount.gt.0)then
         do i=1,icount
           read(10,*,err=9100,end=9100) atemp1,atemp2
           call lowcase(atemp1)
           call lowcase(atemp2)
           do j=1,npar1
             if(atemp1.eq.apar1(j)) go to 170
           end do
           go to 9100
170        continue
           do k=1,npar1
             if(atemp2.eq.apar1(k)) go to 180
           end do
           go to 9100
180        continue
           if(itrans1(j).ne.-1) go to 9100
           if(itrans1(k).lt.0) go to 9100
           itrans1(j)=-k
         end do
       end if

       do
         read(10,'(a)',err=9100,end=200) cline
         call lowcase(cline)
         if(index(cline,'* observation data').ne.0) go to 220
       end do
200    call addquote(pestfile1,afile)
       write(errmsg,210) trim(afile)
210    format('Cannot find "observation data" section of file ',a,'.')
       go to 9890
220    do i=1,nobs1
         read(10,*,err=9100,end=9100) aobs1(i)
         call lowcase(aobs1(i))
       end do
       close(unit=10)

       nespar1=0
       do i=1,npar1
         if(itrans1(i).ge.0)nespar1=nespar1+1
       end do
       if(nespar1.eq.0)then
         call addquote(pestfile1,afile)
         write(errmsg,231) trim(afile)
231      format('There are no adjustable parameters in file ',a,'.')
         go to 9890
       end if

       call addquote(pestfile1,afile)
       write(6,230) trim(afile)
230    format(' - file ',a,' read ok.')

! -- The Jacobian matrix file is read.

       read(15,err=9250,end=9250)itemp1,itemp2
       if(abs(itemp1).ne.nespar1) go to 9200
       bn=abs(itemp2)
       if(itemp1.lt.0)then
         continue
       else
         call addquote(jcofile1,afile)
         write(errmsg,240) trim(afile)
240      format('Jacobian matrix file ',a,' was not produced by ',
     +   'version 8 or later of PEST; use program JCOTRANS to ',
     +   'convert this to new format.')
         go to 9890
       end if

       allocate(x1(nobs1,nespar1),stat=ierr)
       if(ierr.ne.0)then
         write(errmsg,125)
         go to 9890
       end if

       do i=1,nespar1
         do j=1,nobs1
           x1(j,i)=0.0d0
         end do
       end do
       read(15,err=9250,end=9250)icount
       do i=1,icount
         read(15,err=9250,end=9250) j,dtemp
         ies=(j-1)/bn+1
         irow=j-(ies-1)*bn
         if(irow.le.nobs1) x1(irow,ies)=dtemp
       end do
       do i=1,npar1
         if(itrans1(i).lt.0) cycle
         read(15,err=9250,end=9250) aapar
         if(aapar.ne.apar1(i)) go to 9200
       end do
       do i=1,nobs1
         read(15,err=9250,end=9250) aaobs
         call lowcase(aaobs)
         if(aaobs.ne.aobs1(i)) go to 9200
       end do
       close(unit=15)
       call addquote(jcofile1,afile)
       write(6,230) trim(afile)

! -- The second PEST control file is read and its Jacobian file opened, ready
!    for writing.

#ifdef LAHEY
#ifdef UNIX
       open(unit=15,file=jcofile2,form='unformatted',
     + iostat=ierr)
#else
       open(unit=15,file=jcofile2,form='unformatted',
     + access='transparent',iostat=ierr)
#endif
#else
       open(unit=15,file=jcofile2,form='unformatted',
     + iostat=ierr)
#endif
       if(ierr.ne.0)then
         call addquote(jcofile2,afile)
         write(errmsg,250) trim(afile)
250      format('Cannot open Jacobian matrix file ',a,' for output.')
         go to 9890
       end if

! -- Data is read from the second PEST control file.

       open(unit=10,file=pestfile2,status='old',iostat=ierr)
       if(ierr.ne.0)then
         call addquote(pestfile2,afile)
         write(errmsg,110) trim(afile)
         call write_message(errmsg)
         stop
       end if

       do i=1,3
         read(10,*,err=9300,end=9300)
       end do
       read(10,*,err=9300,end=9300) npar2,nobs2,npargp2,nprior2
       allocate(apar2(npar2),pval2(npar2),itrans2(npar2),
     + aobs2(nobs2),scale2(npar2),offset2(npar2),stat=ierr)
       if(ierr.ne.0)then
         write(errmsg,125)
         go to 9890
       end if
       do
         read(10,'(a)',err=9300,end=330) cline
         if(index(cline,'* parameter data').ne.0) go to 350
       end do
330    continue
       call addquote(pestfile2,afile)
       write(errmsg,140) trim(afile)
       go to 9890
350    do i=1,npar2
         read(10,*,err=9300,end=9300) apar2(i),atemp,atemp1,pval2(i),
     +   dtemp,dtemp,atemp1,scale2(i),offset2(i)
         call lowcase(apar2(i))
         call lowcase(atemp)
         if(atemp.eq.'log')then
           itrans2(i)=1
         else if(atemp.eq.'none')then
           itrans2(i)=0
         else if(atemp.eq.'fixed')then
           itrans2(i)=-100000
         else if(atemp.eq.'tied')then
           itrans2(i)=-1
         else
           go to 9300
         end if
       end do
       icount=0
       do i=1,npar2
         if(itrans2(i).eq.-1)icount=icount+1
       end do
       if(icount.gt.0)then
         do i=1,icount
           read(10,*,err=9300,end=9300) atemp1,atemp2
           call lowcase(atemp1)
           call lowcase(atemp2)
           do j=1,npar2
             if(atemp1.eq.apar2(j)) go to 370
           end do
           go to 9300
370        continue
           do k=1,npar2
             if(atemp2.eq.apar2(k)) go to 380
           end do
           go to 9300
380        continue
           if(itrans2(j).ne.-1) go to 9300
           if(itrans2(k).lt.0) go to 9300
           itrans2(j)=-k
         end do
       end if

       do
         read(10,'(a)',err=9300,end=400) cline
         call lowcase(cline)
         if(index(cline,'* observation data').ne.0) go to 420
       end do
400    call addquote(pestfile2,afile)
       write(errmsg,210) trim(afile)
       go to 9890
420    do i=1,nobs2
         read(10,*,err=9300,end=9300) aobs2(i)
         call lowcase(aobs2(i))
       end do

       close(unit=10)
       call addquote(pestfile2,afile)
       write(6,230) trim(afile)

       aprefix='JCO2JCO cannot write second JCO file: '

C -- Dimensions of the second derivatives matrix are evaluated.

       nespar2=0
       do i=1,npar2
         if(itrans2(i).ge.0) nespar2=nespar2+1
       end do
       if(nespar2.gt.nespar1)then
         write(errmsg,450) trim(aprefix)
450      format(a,' second PEST control file has more ',
     +   'adjustable parameters than first.')
         go to 9890
       end if
       nobs3=nobs2
       allocate(x3(nobs3,nespar1),x2(nobs3,nespar2),stat=ierr)
       if(ierr.ne.0)then
         write(errmsg,125)
         go to 9890
       end if

C -- Observations in the first file are re-ordered (and some left out if necessary).

       ii=1
       do iobs=1,nobs2
         aaobs=aobs2(iobs)
         do i=ii,nobs1
           if(aobs1(i).eq.aaobs) go to 600
         end do
         if(ii.ne.1)then
           do i=1,ii-1
             if(aobs1(i).eq.aaobs) go to 600
           end do
         end if
         write(errmsg,590) trim(aprefix),trim(aaobs)
590      format(a,' observation "',a,'" is cited in second ',
     +   'PEST control file but not in first.')
         go to 9890
600      ii=i
         do i=1,nespar1
           x3(iobs,i)=x1(ii,i)
         end do
         ii=ii+1
       end do

! -- All derivatives in first file are converted to natural derivatives.

       ies=0
       do i=1,npar1
         if(itrans1(i).lt.0) cycle
         ies=ies+1
         if(itrans1(i).eq.1)then
           dtemp1=pval1(i)
           if(dtemp1.le.0.0)then
             call addquote(pestfile1,afile)
             write(errmsg,460) trim(apar1(i)),trim(afile)
460          format('Zero or negative log-transformed parameter ',
     +       '("',a,'") in PEST control file ',a,'.')
             go to 9890
           end if
           dtemp1=ilogfac/dtemp1
           do j=1,nobs3
             x3(j,ies)=x3(j,ies)*dtemp1
           end do
         end if
       end do

! -- Some issues with respect to scale are worked out.

       do i=1,npar1
         if((itrans1(i).lt.0).and.(itrans1(i).gt.-100000))then
           aspar1=apar1(i)
           do j=1,npar2
             if(apar2(j).eq.aspar1)then
               if(abs(scale1(i)-scale2(j)).gt.
     +           1.0e-6*abs(scale1(i)+scale2(j)))then
                 write(errmsg,461) trim(aprefix),trim(aspar1)
461              format(a,' parameter "',a,'" is tied ',
     +           'parameter in first PEST control file but has a ',
     +           'different scale in second PEST control file.')
                 go to 9890
               end if
               go to 462
             end if
           end do
         end if
462      continue
       end do
       do i=1,npar2
         if((itrans2(i).lt.0).and.(itrans2(i).gt.-100000))then
           aspar2=apar2(i)
           do j=1,npar1
             if(apar1(j).eq.aspar2)then
               if(abs(scale2(i)-scale1(j)).gt.
     +           1.0e-6*abs(scale2(i)+scale1(j)))then
                 write(errmsg,459) trim(aprefix),trim(aspar2)
459              format(a,' parameter "',a,'" is tied ',
     +           'parameter in second PEST control file but has a ',
     +           'different scale in first PEST control file.')
                 go to 9890
               end if
               go to 458
             end if
           end do
         end if
458      continue
       end do

! -- The second Jacobian matrix file is built.

       ies=0
       do i=1,npar2
         if(itrans2(i).lt.0)cycle
         ies=ies+1
         aapar=apar2(i)
         do jj=1,npar1
           if(apar1(jj).eq.aapar) go to 620
         end do
         write(errmsg,431) trim(aprefix),trim(aapar)
431      format(a,' parameter "',a,'" is adjustable in second ',
     +   'PEST control file but is not found in first.')
         go to 9890
620      continue
         if(itrans1(jj).lt.0)then
           write(errmsg,430) trim(aprefix),trim(aapar)
430        format(a,' parameter "',a,'" is adjustable in second ',
     +     'PEST control file but not in first.')
           go to 9890
         end if
         dd2=pval2(i)
         dd1=pval1(jj)
         if(dd1.eq.dd2)then
           continue
         else
           if(abs(dd1-dd2).gt.abs(1.0d-6*(dd1+dd2)))then
             rstemp2=dd2*scale2(i)
             rstemp1=dd1*scale1(jj)
             if(abs(rstemp2-rstemp1).gt.abs(1.0d-6*(rstemp2+rstemp1)))
     +       idifflag=idifflag+1
           end if
         end if
         if(scale1(jj).eq.scale2(i))then
           continue
         else
           if(abs(scale1(jj)-scale2(i)).gt.1.0d-6*
     +        abs(scale1(jj)+scale2(i))) then
              do jjj=1,npar1
                if(itrans1(jjj).eq.-jj)then
                  write(errmsg,442) trim(aprefix),trim(apar1(jj))
442               format(a,' parameter "',a,'" has different scale ',
     +            'in second PEST control file to that in first. ',
     +            'This is only allowed if no parameter is tied ',
     +            'to this parameter in either file.')
                  go to 9890
                end if
              end do
              do jjj=1,npar2
                if(itrans2(jjj).eq.-i)then
                  write(errmsg,442) trim(aprefix),trim(apar1(jj))
                  go to 9890
                end if
              end do
              if((itrans1(jj).ne.0).or.(itrans2(i).ne.0))then
                write(errmsg,446) trim(aprefix),trim(apar1(jj))
446             format(a,' parameter "',a,'" has a different scale ',
     +          'in second PEST control file from that in first. ',
     +          'This is only allowed if the parameter is not log ',
     +          'transformed, tied or fixed in either file. If ',
     +          'appropriate, ',
     +          'consider running JCO2JCO more than once with only ',
     +          'incremental changes to PEST control file on each ',
     +          'occasion.')
                go to 9890
             end if
           end if
           if((scale1(jj).eq.0.0d0).or.(scale2(i).eq.0.0d0))then
             write(errmsg,444) trim(aprefix),trim(apar1(jj))
444          format(a,' parameter "',a,'" has a zero scale in ',
     +       'either first PEST control file or second PEST control ',
     +       'file or both.')
             go to 9890
           end if
         end if
         if(offset1(jj).eq.offset2(i))then
           continue
         else
           if(abs(offset1(jj)-offset2(i)).gt.1.0d-6*
     +        abs(offset1(jj)+offset2(i))) then
              write(errmsg,443) trim(aprefix),trim(apar1(jj))
443           format(a,' parameter "',a,'" has different offset ',
     +        'in second PEST control file to that in first.')
              go to 9890
           end if
         end if

         do j=1,npar1
           if(-itrans1(j).eq.jj)then
             do k=1,npar2
               if(apar2(k).eq.apar1(j))then
                 if(-itrans2(k).ne.i)then
                   write(errmsg,630) trim(aprefix),trim(apar1(j)),
     +             trim(apar1(jj))
630                format(a,' parameter "',a,'" is tied to parameter "',
     +             a,'" in first PEST control file, but not in ',
     +             'second.')
                   go to 9890
                 else
                   go to 640
                 end if
               end if
             end do
             write(errmsg,630) trim(aprefix),trim(apar1(j)),
     +       trim(apar1(jj))
             go to 9890
           end if
640        continue
         end do
         kes=0
         do j=1,jj
           if(itrans1(j).lt.0) cycle
           kes=kes+1
         end do
         if(abs(scale1(jj)-scale2(i)).gt.1.0d-6*
     +      abs(scale1(jj)+scale2(i))) then
           scalefac=scale2(i)/scale1(jj)
           do iobs=1,nobs3
             x2(iobs,ies)=x3(iobs,kes)*scalefac
           end do
         else
           do iobs=1,nobs3
             x2(iobs,ies)=x3(iobs,kes)
           end do
         end if
         jes=0
         do j=1,npar2
           if(-itrans2(j).eq.i)then
             aapar=apar2(j)
             jes=0
             do k=1,npar1
               if(itrans1(k).ge.0)jes=jes+1
               if(apar1(k).eq.aapar)then
                 if(-itrans1(k).eq.jj) then
                   if(pval1(jj).eq.0.0d0) go to 9100
                   ratio1=pval1(k)/pval1(jj)
                   if(pval2(i).eq.0.0d0) go to 9300
                   ratio2=pval2(j)/pval2(i)
                   if(ratio1-ratio2.gt.1.0e-6* ratio1)then
                     write(errmsg,669)trim(apar2(i))
669                  format('Parameters tied to parameter "',a,
     +               '" in two PEST control files are not tied in ',
     +               'the same ratio.')
                     go to 9890
                   end if
                   continue
                 else if(itrans1(k).eq.-100000)then
                   write(errmsg,670) trim(aprefix),trim(aapar)
670                format(a,' parameter "',a,'" is fixed in first ',
     +             'PEST control file but not in second.')
                   go to 9890
                 else
                   ratio=pval2(j)/pval2(i)
                   do iobs=1,nobs3
                     x2(iobs,ies)=x2(iobs,ies)+x3(iobs,jes)*ratio
                   end do
                 end if
                 go to 680
               end if
             end do
             write(errmsg,690) trim(aprefix),trim(aapar)
690          format(a,' parameter "',a,'"  is in second PEST control ',
     +       'file but not in first.')
             go to 9890
680          continue
           end if
         end do
       end do

C -- Log-transformation is undertaken as necessary.

       ies=0
       do i=1,npar2
         if(itrans2(i).lt.0) cycle
         ies=ies+1
         if(itrans2(i).eq.1)then
           if(pval2(i).le.0.0d0)then
             write(errmsg,695) trim(apar2(i)),trim(pestfile2)
695          format('Parameter "',a,'" cannot be log-transformed in ',
     +       'PEST control file "',a,
     +       '" as it has a non-positive value.')
             go to 9890
           end if
           dtemp=logfac*pval2(i)
           do iobs=1,nobs3
             x2(iobs,ies)=x2(iobs,ies)*dtemp
           end do
         end if
       end do

C -- The Jacobian matrix is written.

!#ifdef LAHEY
!#ifdef UNIX
!       open(unit=15,file=jcofile2,form='unformatted',
!C     + action='write',err=9400)
!     + err=9400)    ! caused a problem with Lahey lf95 unix
!#else
!       open(unit=15,file=jcofile2,form='unformatted',
!     + access='transparent',action='write',err=9400)
!#endif
!#else
!       open(unit=15,file=jcofile2,form='unformatted',
!C     + action='write',err=9400)    ! caused a problem with Lahey lf95 unix
!     + err=9400)
!#endif

       write(15,err=9400) -nespar2,-nobs2
       icount=0
       do i=1,nespar2
          do j=1,nobs2
            if(x2(j,i).ne.0.0d0) icount=icount+1
          end do
       end do
       write(15,err=9400) icount
       do i=1,nespar2
         do j=1,nobs2
           if(x2(j,i).ne.0.0d0)then
             icount=(i-1)*nobs2+j
             write(15,err=9400)icount,x2(j,i)
           end if
         end do
       end do
       do ipp=1,npar2
         if((itrans2(ipp).eq.0).or.(itrans2(ipp).eq.1))then
            write(15,err=9400) apar2(ipp)
         end if
       end do
       do iobs=1,nobs2
         write(15,err=9400) aobs2(iobs)
       end do
       close(unit=15)
       call addquote(jcofile2,afile)
       write(6,490) trim(afile)
490    format(' - file ',a,' written ok.')
       if(idifflag.ne.0)then
         write(6,495)
495      format(/,' Warning: at least one adjustable parameter from ',
     +   'second PEST control file has ',/,
     +   ' different value from that of parameter of same name in ',
     +   'first PEST control file.')
       end if

       go to 9900

9000   continue

       write(6,9010)
9010   format(/,' JCO2JCO is run using the command:',/)
       write(6,9020)
9020   format('     jco2jco casename1 casename2',/)
       write(6,9030)
9030   format(' where',/)
       write(6,9050)
9050   format('     casename1 pertains to a case for which a ',
     + 'JCO file exists, and')
       write(6,9060)
9060   format('     casename2 pertains to a case for which a ',
     + 'JCO file must be written.')
       go to 9999

9100   call addquote(pestfile1,afile)
       write(errmsg,9110) trim(afile)
9110   format('Error reading PEST control file ',a,'; check this ',
     + 'file with PESTCHEK.')
       go to 9890

9200   continue
       call addquote(jcofile1,afile)
       call addquote(pestfile1,bfile)
       write(errmsg,9210) trim(afile),trim(bfile)
9210   format('Jacobian matrix file ',a,' is incompatible with PEST ',
     + 'control file ',a,'.')
       go to 9890
9250   call addquote(jcofile1,afile)
       write(errmsg,9260) trim(afile)
9260   format('Error reading Jacobian matrix file ',a,'.')
       go to 9890
9300   call addquote(pestfile2,afile)
       write(errmsg,9110) trim(afile)
       go to 9890
9400   call addquote(jcofile2,afile)
       write(errmsg,9410) trim(afile)
9410   format('Cannot write to file ',a,'.')
       go to 9890

9890   call write_message(errmsg)
9900   continue

       deallocate (itrans1,itrans2,pval1,pval2,x1,x2,apar1,aobs1,
     + apar2,aobs2,x3,scale1,scale2,offset1,offset2,stat=ierr)

9999   continue
       end


        subroutine linspl(ifail,num,lw,rw,cline)

c -- Subroutine LINSPL splits a line into whitespace-separated substrings.

        integer ifail,nw,nblc,j,i
        integer num
        integer lw(num),rw(num)
        character*(*) cline

        ifail=0
        nw=0
        nblc=len_trim(cline)
        if(nblc.eq.0) then
          ifail=1
          return
        end if
        j=0
5       if(nw.eq.num) return
        do 10 i=j+1,nblc
        if((cline(i:i).ne.' ').and.(cline(i:i).ne.',')
     +  .and.(ichar(cline(i:i)).ne.9)) go to 20
10      continue
        ifail=1
        return
20      nw=nw+1
        lw(nw)=i
        do 30 i=lw(nw)+1,nblc
        if((cline(i:i).eq.' ').or.(cline(i:i).eq.',')
     +  .or.(ichar(cline(i:i)).eq.9)) go to 40
30      continue
        rw(nw)=nblc
        if(nw.lt.num) ifail=1
        return
40      rw(nw)=i-1
        j=rw(nw)
        go to 5

        end



        subroutine lowcase(astrng)

C -- Subroutine LOWCASE converts a string to lower case.

        integer i,j
        character*(*) astrng

        do 10 i=1,len_trim(astrng)
        j=ichar(astrng(i:i))
        if((j.ge.65).and.(j.le.90)) astrng(i:i)=char(j+32)
10      continue
        return
        end


        subroutine addquote(afile,aqfile)

C -- Subroutine ADDQUOTE adds quotes to a filename if it has a space in it.

        character*(*) afile,aqfile
        integer nbb

        if(index(trim(afile),' ').eq.0)then
          aqfile=afile
        else
          aqfile(1:1)='"'
          aqfile(2:)=trim(afile)
          nbb=len_trim(aqfile)+1
          aqfile(nbb:nbb)='"'
        end if

        return
        end


        subroutine write_message(errmsg)

C -- Subroutine write_message formats and writes a message to the screen.

        implicit none

        integer  junit
        integer  jend,i,nblc,ierr,j
        character*(*) errmsg

        junit=6
        write(junit,*)

        j=0
        nblc=len_trim(errmsg)
5       jend=j+78
        if(jend.ge.nblc) go to 100
        do i=jend,j+1,-1
          if(errmsg(i:i).eq.' ') then
            write(junit,50,iostat=ierr) errmsg(j+1:i)
50          format(1x,a)
            j=i
            go to 5
          end if
        end do
        write(junit,50,iostat=ierr) errmsg(j+1:jend)
        j=jend
        go to 5
100     jend=nblc
        write(junit,50,iostat=ierr) errmsg(j+1:jend)
        write(junit,*)
        return

        end


        subroutine intread(ifail,cline,itemp)

C -- Subroutine INTREAD reads a real number from a string.

        integer ifail
        integer itemp
        character*6 afmt
        character*(*) cline

        ifail=0
        afmt='(i   )'
        write(afmt(3:5),'(i3)') len(cline)
        read(cline,afmt,err=100) itemp
        return

100     ifail=1
        return
        end


