MODULE PARALLEL_RUNS

! -- Dimension variables

    INTEGER  :: MAXPLLRUNS    ! Maximum number of parallel runs
    INTEGER  :: NSLAVE        ! Number of slaves
    INTEGER  :: NUMINFILE     ! Number of model input files
    INTEGER  :: NUMOUTFILE    ! Number of model output files

! -- Control variables

    INTEGER            :: PLLMODE      ! Parallel run configuration
    INTEGER            :: IWAIT        ! Wait period in 100ths of a second.
    INTEGER            :: INI_CALL=0          ! Flag indicating subroutine call.
    INTEGER            :: MODINOUTFILE_CALL=0 ! Flag indicating subroutine call.
    INTEGER            :: SLAVE_INFO_CALL=0   ! Flag indicating subroutine call.
    INTEGER            :: INI_COMM_CALL=0     ! Flag indicating subroutine call.
    INTEGER            :: DO_RUNS_CALL=0      ! Flag indicating subroutine call.
    INTEGER            :: NNRUN=0             ! Total number of model runs.

! -- Output variables

    INTEGER             :: GLOBALUNIT ! Unit no. of global output file
    INTEGER             :: RMRUNIT   ! Unit number for run management record file
    CHARACTER (LEN=256) :: RMRFILE   ! Name of run management record file

! -- Utility variables

#ifdef UNIX
    CHARACTER (LEN=1)   :: SLASH_CHAR='/'
#else
    CHARACTER (LEN=1)   :: SLASH_CHAR='\'
#endif
    CHARACTER (LEN=256) :: AFILE     ! Used for storing temporary filenames.
    CHARACTER (LEN=256) :: SINFLE    ! Current model input file
    CHARACTER (LEN=256) :: SOUFLE    ! Current model output file

! -- Variables used in run assignment and slave control.

    INTEGER, ALLOCATABLE      :: IDET(:)    ! Slave detection indicator.
    INTEGER, ALLOCATABLE      :: ISTATS(:)  ! Status of each slave
    INTEGER, ALLOCATABLE      :: ISTATR(:)  ! Status of each model run
    INTEGER, ALLOCATABLE      :: IRUNTME(:) ! Slave run time in seconds.
    INTEGER, ALLOCATABLE      :: ISTRTME(:) ! Start time of each run
    INTEGER, ALLOCATABLE      :: JRUN(:)    ! Run currently undertaken by a slave
    INTEGER, ALLOCATABLE      :: ITRIAL(:)  ! No. of trials for a run
    CHARACTER (LEN=30),ALLOCATABLE  :: ASLAVE(:)  ! Name of slave.
    CHARACTER (LEN=200),ALLOCATABLE :: ASLDIR(:)  ! Slave directory as seen my master.

! -- Communications files

    INTEGER              ::  TEMPUNIT
    INTEGER              ::  TEMPUNIT2
    CHARACTER (LEN=200)  ::  MREADFLE='master##.rdy'
    CHARACTER (LEN=200)  ::  SREADFLE='slave##.rdy'
    CHARACTER (LEN=200)  ::  PREADFLE='param##.rdy'
    CHARACTER (LEN=200)  ::  OREADFLE='observ##.rdy'
    CHARACTER (LEN=200)  ::  SFINFLE ='slave##.fin'

! -- Model input/output files

    DOUBLE PRECISION, ALLOCATABLE :: RUNPARVAL(:) ! Values to write to model input files
    CHARACTER (LEN=200)  :: MODEL_INPUT_FILE(:)   ! Model input file
    CHARACTER (LEN=200)  :: MODEL_OUTPUT_FILE(:)  ! Model output file
    CHARACTER (LEN=200),ALLOCATABLE  :: SCOM(:)   ! Model command used by slave.

! -- Variables pertaining to time

    CHARACTER (LEN=11)   :: RESULT
    INTEGER              :: M1, D1, Y1, H1, MIN1
    REAL                 :: SS1


  PRIVATE

  PUBLIC PLL_INI, PLL_MODINOUTFILE, PLL_SLAVE_INFO, PLL_INI_COMM, PLL_DO_RUNS, &
         PLL_FIN

CONTAINS

  SUBROUTINE PLL_INI(IFAIL,IOUTG,IOUT,JOUT1,JOUT2,MODE,MAXRUNS,NUMSLAVES,NCONTROL,   &
                     NUMODINFILE,NUMODOUTFILE,OUTFILE)

! -- Subroutine PLL_INI receives control and dimension variables from the
!    main program.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       IMPLICIT NONE

       INTEGER, INTENT(OUT)          :: IFAIL
       INTEGER, INTENT(IN)           :: IOUTG ! Unit number for main pgme o/p file.
       INTEGER, INTENT(IN)           :: IOUT  ! Unit number for run mgt rec file.
       INTEGER, INTENT(IN)           :: JOUT1  ! Unit number for temporary files.
       INTEGER, INTENT(IN)           :: JOUT2  ! Unit number for temporary files.
       INTEGER, INTENT(IN)           :: MODE
       INTEGER, INTENT(IN)           :: MAXRUNS
       INTEGER, INTENT(IN)           :: NUMSLAVES
       INTEGER, INTENT(IN)           :: NCONTROL
       INTEGER, INTENT(IN)           :: NUMODINFILE
       INTEGER, INTENT(IN)           :: NUMODOUTFILE
       CHARACTER (LEN=*), INTENT(IN) :: OUTFILE

! -- Initialisation

       IFAIL=0
       INI_CALL=1
       ERRSUB='Error in subroutine EQN_INI:'

! -- Error-checking.

       IF((IOUT.LE.0).OR.(IOUTG.LE.0).OR.(IOUT1.LE.0).OR.(IOUT2.LE.0))THEN
         WRITE(AMESSAGE,10) TRIM(ERRSUB)
10       FORMAT(A,' improper output unit number supplied.')
         GO TO 9900
       END IF
       IF(MODE.NE.1)THEN
         WRITE(AMESSAGE,20) TRIM(ERRSUB)
20       FORMAT(A,' improper parallelisation mode variable.')
         GO TO 9900
       END IF
       IF(MAXRUNS.LE.0)THEN
         WRITE(AMESSAGE,30) TRIM(ERRSUB)
30       FORMAT(A,' maximum number of runs supplied as negative.')
         GO TO 9900
       END IF
       IF(NUMSLAVE.LE.0)THEN
         WRITE(AMESSAGE,40) TRIM(ERRSUB)
40       FORMAT(A,' maximum number of slaves supplied as negative.')
         GO TO 9900
       END IF
       IF(NCONTROL.LE.1)THEN
         WRITE(AMESSAGE,50) TRIM(ERRSUB)
50       FORMAT(A,' dimension of control variable array must be ',  &
         '1 or greater.')
         GO TO 9900
       END IF
       IF((NUMMONDINFILE.LE.0).OR.(NUMMODOUTFILE.LE.0))THEN
         WRITE(AMESSAGE,60) TRIM(ERRSUB)
60       FORMAT(A,' number of model input files and/or number ', &
         'of model output files supplied as zero.')
         GO TO 9900
       END

! -- Values are assigned to model variables.

       TEMPUNIT=JOUT1
       TEMPUNIT2=JOUT2
       RMRUNIT=IOUT
       GLOBALUNIT=IOUTG
       RMRFILE=OUTFILE
       PLLMODE=MODE
       NSLAVE=NUMSLAVES
       MAXPLLRUNS=MAXRUNS
       NUMINFILE=NUMODINFILE
       NUMOUTFILE=NUMODOUTFILE
       IWAIT=NINT(CONTROL(1)*100.0)
       IF(IWAIT.LE.0)THEN
         WRITE(AMESSAGE,70) TRIM(ERRSUB)
70       FORMAT(A,' WAIT variable supplied as negative.')
         GO TO 9900
       END IF

! -- Memory is allocated.

       ALLOCATE(IDET(NSLAVE),IRUNTME(NSLAVE),ASLAVE(NSLAVE),ALSDIR(NSLAVE), &
       SCOM(NSLAVE),ISTATS(NSLAVE),ISTATR(NRUN),ISTRTME(NSLAVE),JRUN(NSLAVE), &
       ITRIAL(NRUN),STAT=IERR)

       IF(IERR.NE.0)THEN
         WRITE(AMESSAGE,110) TRIM(ERRSUB)
110      FORMAT(A,' cannot allocate arrays to hold slave data.')
         GO TO 9900
       END IF
       ALLOCATE(MODEL_INPUT_FILE(NUMINFILE),MODEL_OUTPUT_FILE(NUMOUTFILE), &
       STAT=IERR)
       IF(IERR.NE.0)THEN
         WRITE(AMESSAGE,110) TRIM(ERRSUB)
         GO TO 9900
       END IF
       IDET=0
       IRUNTME=0
       ASLAVE=' '
       ALSDIR=' '
       SCOM='Not yet known.'
       ISTATS=0
       ISTATR=0
       ISTRTME=0
       JRUN=0
       ITRIAL=0

       GO TO 9999

9900   IFAIL=1

9999   RETURN
       END SUBROUTINE PLL_INI


       SUBROUTINE PLL_MODINOUTFILE(IFAIL,MODINFILE,MODOUTFILE)

       IMPLICIT NONE
       INTEGER, INTENT(OUT)           :: IFAIL
       CHARACTER (LEN=*), INTENT(IN)  :: MODINFILE(NUMINFILE)
       CHARACTER (LEN=*), INTENT(IN)  :: MODOUTFILE(NUMOUTFILE)

       INTEGER                        :: I

       IFAIL=0
       MODINOUTFILE_CALL=1
       IF(INI_CALL.EQ.0)THEN
10       FORMAT('Programming error: call to PLL_MODINOUTFILE must be made only ', &
         'after call to PLL_INI.')
         IFAIL=1
         RETURN
       END

       DO I=1,NUMINFILE
         MODEL_INPUT_FILE(I)=MODINFILE(I)
       END DO
       DO I=1,NUMOUTFILE
         MODEL_OUTPUT_FILE(I)=MODOUTFILE(I)
       END DO

       RETURN
       END




       SUBROUTINE PLL_SLAVE_INFO(IFAIL,ISLAVE,SLAVENAME,DIRECTORY,RUNTIME)

! -- Subroutine PLL_SLAVE_INFO receives information pertaining to operative
!    slaves.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       IMPLICIT NONE

       INTEGER, INTENT(OUT)          :: IFAIL     ! Failure indicator
       INTEGER, INTENT(IN)           :: ISLAVE    ! Slave for which information is provided
       CHARACTER (LEN=*), INTENT(IN) :: SLAVENAME ! Name of slave.
       CHARACTER (LEN=*), INTENT(IN) :: DIRECTORY ! Slave working directory as seen by master.
       REAL, INTENT(IN)              :: RUNTIME   ! Estimated runtime in seconds.

       IFAIL=0
       SLAVE_INFO_CALL=1
       IF(MODINOUTFILE_CALL.EQ.0)THEN
10       FORMAT('Programming error: call to PLL_SLAVE_INFO must be made only ', &
         'after call to PLL_MODINOUTFILE.')
         IFAIL=1
         RETURN
       END

       ERRSUB='Error in subroutine EQN_INI:'

       IF(ISLAVE.LE.0).OR.(ISLAVE.GT.NSLAVE))THEN
         WRITE(AMESSAGE,40) TRIM(ERRSUB)
40       FORMAT(A,' slave number out of range.')
         GO TO 9900
       END IF
       ASLAVE(ISLAVE)=SLAVENAME
       IRUNTME(ISLAVE)=RUNTIME
       N=LEN_TRIM(DIRECTORY)
       IF(DIRECTORY(N:N).NE.SLASH_CHAR)THEN
         N=N+1
         DIRECTORY(N:N)=SLASH_CHAR
       END IF
       ASLDIR(ISLAVE)=DIRECTORY

! -- The run management record file is opened.
! -- If the compiler permits it, open this with zero blocksize.

#ifdef LF90
       OPEN(UNIT=RMRUNIT,FILE=RMRFILE,ACTION='WRITE',   &
       BLOCKSIZE=1,IOSTAT=IERR)
#else
       OPEN(UNIT=RMRUNIT,FILE=RMRFILE,ACTION='WRITE',IOSTAT=IERR)
#endif
       IF(IERR.NE.0)THEN
         CALL ADDQUOTE(RMRFILE,AFILE)
         WRITE(AMESSAGE,410) TRIM(RMRFILE)
410      FORMAT('Cannot open run management file ',a,'.')
         GO TO 9900
       END IF

C -- Some information is next recorded on the run manager record file.

       WRITE(RMRUNIT,430)
430    FORMAT(T25,'PARALLEL RUN MANAGEMENT RECORD')
       CALL WRITE_SLAVE_DATA_1(JFAIL)
       IF(JFAIL.NE.0) GO TO 9900

       GO TO 9999

9900   CONTINUE
       IFAIL=1

9999   CONTINUE
       RETURN

       END SUBROUTINE PLL_SLAVE_INFO




       SUBROUTINE PLL_INI_COMM(IFAIL)

! -- Subroutine PLL_INI_COMM initiates communication between the master program
!    and its slaves.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       IMPLICIT NONE
       INTEGER, INTENT(OUT)     :: IFAIL


       INTEGER            :: ISLAVE,IERR,NUMERR,N,JFAIL,NDET
       CHARACTER (LEN=10) :: ATEMP

! -- Initialisation

       IFAIL=0
       INI_COMM_CALL=1
       IF(SLAVE_INFO_CALL.EQ.0)THEN
10       FORMAT('Programming error: call to PLL_INI_COMM must be made only ', &
         'after call to PLL_SLAVE_INFO.')
         IFAIL=1
         RETURN
       END
       NUMERR=0

! -- It is established whether files can be read/written in each slave subdirectory.

       DO ISLAVE=1,NSLAVE
         AFILE=TRIM(ASLDIR(ISLAVE))//'p###.##'
         OPEN(UNIT=TEMPUNIT,FILE=AFILE,IOSTAT=IERR)
           IF(IERR.NE.0) THEN
           WRITE(6,150)
150        FORMAT(' Parallel run management warning:-')
           WRITE(6,160) TRIM(ASLDIR(ISLAVE))
160        FORMAT(' Cannot open a file in subdirectory ',A)
           NUMERR=NUMERR+1
           GO TO 400
         END IF
         WRITE(TEMPUNIT,'(A)',IOSTAT=IERR) ' Testing'
         IF(IERR.NE.0) THEN
           WRITE(AMESSAGE,170) TRIM(ERRSUB),TRIM(ASLDIR(ISLAVE))
170        FORMAT(A,' Cannot write to a file in subdirectory ',A)
           GO TO 9891
         END IF
         CLOSE(UNIT=TEMPUNIT,IOSTAT=IERR)
         IF(IERR.NE.0) THEN
           WRITE(AMESSAGE,190) TRIM(ERRSUB),TRIM(ASLDIR(ISLAVE))
190        FORMAT(A,' Cannot close a file in subdirectory ',A)
           GO TO 9891
         END IF
         AFILE=TRIM(ASLDIR(ISLAVE))//'p###.##'
         OPEN(UNIT=TEMPUNIT,FILE=AFILE,STATUS='OLD',IOSTAT=IERR)
         IF(IERR.NE.0) THEN
           WRITE(6,150)
           WRITE(6,160) TRIM(ASLDIR(ISLAVE))
           NUMERR=NUMERR+1
           GO TO 400
         END IF
         READ(TEMPUNIT,'(A)',IOSTAT=IERR) ATEMP
         IF(IERR.NE.0) THEN
           WRITE(AMESSAGE,210) TRIM(ERRSUB),TRIM(ASLDIR(ISLAVE))
210        FORMAT(A,' cannot read from a file in subdirectory ',A)
           GO TO 9891
         END IF
         CLOSE(UNIT=TEMPUNIT,IOSTAT=IERR)
         IF(IERR.NE.0) THEN
           WRITE(AMESSAGE,190) TRIM(ERRSUB),TRIM(ASLDIR(ISLAVE))
           GO TO 9891
         END IF
400    CONTINUE

C -- Next we see if the slaves have started.

	DO 500 ISLAVE=1,NSLAVE
          AFILE=TRIM(ASLDIR(ISLAVE))//SREADFLE
          CALL DELFILE(JFAIL,-50,AFILE)
          IF(JFAIL.NE.0) GO TO 9891
          AFILE=TRIM(ASLDIR(ISLAVE))//OREADFLE
          CALL DELFILE(JFAIL,-50,AFILE)
          IF(JFAIL.NE.0) GO TO 9891
          AFILE=TRIM(ASLDIR(ISLAVE))//PREADFLE
          CALL DELFILE(JFAIL,-50,AFILE)
          IF(JFAIL.NE.0) GO TO 9891
          AFILE=TRIM(ASLDIR(ISLAVE))//SFINFLE
          CALL DELFILE(JFAIL,-50,AFILE)
          IF(JFAIL.NE.0) GO TO 9891
          AFILE=TRIMASLDIR(ISLAVE))//MREADFLE
#ifdef LF90
          OPEN(UNIT=TEMPUNIT,FILE=AFILE(1:NBLNK(AFILE)),    &
          ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,IOSTAT=IERR)
#else
          OPEN(UNIT=TEMPUNIT,FILE=AFILE(1:NBLNK(AFILE)),IOSTAT=IERR)
#endif
          IF(IERR.EQ.0)THEN
            WRITE(TEMPUNIT,'(I10)',ERR=9200) IWAIT
            CALL CLOSEFILE(JFAIL,79,2,AFILE)
            IF(JFAIL.NE.0) GO TO 9891
          END IF
500	CONTINUE

	CALL WAIT(IIWAIT)
	CALL WAIT(IIWAIT)

520     CONTINUE
        WRITE(6,525)
525     FORMAT(' TESTING WHETHER SLAVES ARE ALIVE .....',/)
        N=0
        NDET=0
530     DO 600 ISLAVE=1,NSLAVE
          IF(IDET(ISLAVE).EQ.0) THEN
            AFILE=TRIM(ASLDIR(ISLAVE))//SREADFLE
#ifdef LF90
            OPEN(UNIT=TEMPUNIT,FILE=AFILE(1:NBLNK(AFILE)),STATUS='OLD',  &
            ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,IOSTAT=IERR)
#else
            OPEN(UNIT=TEMPUNIT,FILE=AFILE(1:NBLNK(AFILE)),STATUS='OLD',  &
            IOSTAT=IERR)
#endif
            IF(IERR.EQ.0)THEN
              READ(TEMPUNIT,'(A)') SCOM(ISLAVE)
              CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
              IF(JFAIL.NE.0) GO TO 9891
              WRITE(6,550) TRIM(ASLAVE(ISLAVE))
550           FORMAT(' - slave "',A,'" has been detected.')
              WRITE(RMRUNIT,550) TRIM(ASLAVE(ISLAVE))
              IDET(ISLAVE)=1
              NDET=NDET+1
              AFILE=TRIM(ASLDIR(ISLAVE))//MREADFLE
#ifdef LF90
              OPEN(UNIT=TEMPUNIT,FILE=AFILE,                            &
              ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9200)
#else
              OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9200)
#endif
              WRITE(TEMPUNIT,'(I10)',ERR=9200) IWAIT
              CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
              IF(JFAIL.NE.0) GO TO 9891
            END IF
          END IF
600     CONTINUE
        IF(NDET.LT.NSLAVE) THEN
          N=N+1
          IF(N.GT.10) THEN
            IF(NDET.GT.0)THEN
	      CALL WRITINT(ATEMP,NDET)
	      WRITE(6,630) TRIM(ATEMP)
630           FORMAT(/,' Only ',A,' slaves are alive.')
              WRITE(6,640)
640           FORMAT(' Only these slaves will be used until others ',  &
              'become available.')
              GO TO 9999
            ELSE
              WRITE(AMESSAGE,650)
650           FORMAT('No slaves are alive. Start at least one slave and then ', &
              'restart this program.')
              GO TO 9891
            END IF
          END IF
	  IITEMP=MAX(200,IWAIT*3/NSLAVE)
	  CALL WAIT(IITEMP)
	  GO TO 530
	END IF
	WRITE(6,620)
620	FORMAT(/,' All slaves are alive.',/)

C -- Some further information is written to the run manager file.

        CALL WRITE_SLAVE_DATA_2(JFAIL)
        IF(JFAIL.NE.0) GO TO 9891

	GO TO 9999

9200	CONTINUE
        WRITE(AMESSAGE,9210) TRIM(ERRSUB),TRIM(AFILE)
9210    FORMAT(A,' cannot communicate with file ',A)
	GO TO 9891

9891    IFAIL=1

9999    CONTINUE
        RETURN
	END





       SUBROUTINE PLL_DO_RUNS(IFAIL,NRUNS,NPE,NPT,PRECIS,NOPNT,IPTR,PARVALSETS,  &
                              APAR,PVAL,NW,NCATS,NVEXT,EXTNAM,MVCAT,MCUSE,EXTVAL, &
                              INSTRUCTION)

! -- Subroutine PLL_DO_RUNS supervises the carrying out of model runs.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       IMPLICIT NONE

       INTEGER, INTENT(OUT)    :: IFAIL   ! Failure indicator
       INTEGER, INTENT(IN)     :: NRUNS   ! No. of model runs to carry out
       INTEGER, INTENT(IN)     :: NPE     ! No. of adjustable parameters
       INTEGER, INTENT(IN)     :: NPT     ! Total no. of parameters
       INTEGER, INTENT(IN)     :: PRECIS  ! Precision protocol
       INTEGER, INTENT(IN)     :: NOPNT   ! Decimal point protocol
       INTEGER, INTENT(IN)     :: IPTR(NPE) ! Pointer in PVAL to each adj. param.
       DOUBLE PRECISION, INTENT(IN)  :: PARVALSETS(NRUNS,NPE) ! Values of adj. params
       CHARACTER (LEN=*), INTENT(IN) :: APAR(NPT)   ! Parameter names
       DOUBLE PRECISION, INTENT(IN)  :: PVAL(NPT)   ! Current unperterbed param values
       INTEGER, INTENT(IN)           :: NW(NPT)     ! Parameter word lengths
       INTEGER, INTENT(IN)           :: NCATS  ! Number of categories of model-calculated values
       INTEGER, INTENT(IN)           :: NVEXT  ! number of values to extract
       CHARACTER (LEN=*), INTENT(IN) :: EXTNAM(NVEXT) ! extracted-value names
       CHARACTER (LEN=6), INTENT(IN) :: MCVCAT(NCATS) ! Categories supported by application
       LOGICAL, INTENT(IN)           :: MCVUSE(NCATS) ! Do extractions for this category?
       DOUBLE PRECISION, INTENT(OUT) :: EXTVAL(NRUNS,NVEXT) ! extracted values
       CHARACTER (LEN=*),INTENT(OUT) :: INSTRUCTION ! instruction

       INTEGER            :: ISLAVE,IERR,NUMERR,N,JFAIL,NDET,I_ALLOC_FLAG
       CHARACTER (LEN=10) :: ATEMP

! -- Initialisation

       ERRSUB='Error in subroutine PLL_DO_RUNS:'
       IFAIL=0
       I_ALLOC_FLAG=0
       IF(DO_RUNS_CALL.EQ.0) I_ALLOC_IFLAG=1
       DO_RUNS_CALL=1
       IF(INI_COMM_CALL.EQ.0)THEN
10       FORMAT('Programming error: call to PLL_DO_RUNS must be made only ', &
         'after call to PLL_INI_COMM.')
         IFAIL=1
         RETURN
       END

       MAXTRIAL=4
       IIWAIT=IWAIT
       IIRUN=0
       DO IRUN=1,NRUN
         ISTATR(IRUN)=0
         ITRIAL(IRUN)=0
       END DO
       DO ISLAVE=1,NSLAVE
         JRUN(ISLAVE)=0
       END DO

! -- If this is the first call to PLL_DO_RUNS some memory is allocated.

       IF(NPT.LE.0)THEN
         WRITE(AMESSAGE,20) TRIM(ERRSUB)
20       FORMAT(A,' value for NPT supplied as negative.')
         GO TO 9891
       END IF
       IF(I_ALLOC_FLAG.EQ.0) THEN
         ALLOCATE(RUNPARVAL(NPT),STAT=IERR)
         IF(IERR.NE.0) THEN
           WRITE(AMESSAGE,20) TRIM(ERRSUB)
20         FORMAT(A,' cannot allocate sufficient memory to continue ',  &
           'execution.')
           GO TO 9891
         END IF
       END IF

50     CONTINUE

! -- See if any new slaves have appeared.

       DO ISLAVE=1,NSLAVE
         IF(IDET(ISLAVE).EQ.0) THEN
           AFILE=TRIM(ASLDIR(ISLAVE))//SREADFLE
#ifdef LF90
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,STATUS='OLD',                  &
           ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,IOSTAT=IERR)
#else
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,STATUS='OLD',IOSTAT=IERR)
#endif
           IF(IERR.EQ.0)THEN
             READ(TEMPUNIT,'(A)') SCOM(ISLAVE)
             CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
             IF(JFAIL.NE.0) GO TO 9891
             WRITE(RMRUNIT,750) RESULT,TRIM(ASLAVE(ISLAVE))
750          FORMAT(3X,A,':- slave "',a,'" has just been detected.')
             WRITE(6,751) TRIM(ASLAVE(ISLAVE))
751          FORMAT('    Slave ',a,' has just been detected.')
             WRITE(6,*)
             IDET(ISLAVE)=1
             AFILE=TRIM(ASLDIR(ISLAVE))//MREADFLE
#ifdef LF90
             OPEN(UNIT=TEMPUNIT,FILE=AFILE,                   &
             ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9400)
#else
             OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9400)
#endif
             WRITE(TEMPUNIT,'(I10)',ERR=9400) IWAIT
             CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
             IF(JFAIL.NE.0) GO TO 9891
	   END IF
	 END IF
       END DO

! -- Have any slaves from lost runs or re-started slaves re-appeared.

       DO ISLAVE=1,NSLAVE
         IF(ISTATS(ISLAVE).EQ.-999)THEN
           AFILE=TRIM(ASLDIR(ISLAVE))//SREADFLE
#ifdef LF90
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,STATUS='OLD',                  &
           ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,IOSTAT=IERR)
#else
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,STATUS='OLD',IOSTAT=IERR)
#endif
           IF(IERR.EQ.0)THEN
             READ(TEMPUNIT,'(A)') SCOM(ISLAVE)
             CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
             IF(JFAIL.NE.0) GO TO 9891
             AFILE=TRIM(ASLDIR(ISLAVE))//MREADFLE
#ifdef LF90
             OPEN(UNIT=TEMPUNIT,FILE=AFILE,                             &
             ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9400)
#else
             OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9400)
#endif
             WRITE(TEMPUNIT,'(I10)',ERR=9400) IWAIT
             CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
             IF(JFAIL.NE.0) GO TO 9891
             AFILE=TRIM(ASLDIR(ISLAVE))//OREADFLE
             CALL DELFILE(JFAIL,-TEMPUNIT2,AFILE)
             IF(JFAIL.NE.0) GO TO 9891
             ISTATS(ISLAVE)=0
           END IF
         END IF
       END DO

! -- See if it is time to try any "frozen" slaves again.

       DO ISLAVE=1,NSLAVE
         IF(IDET(ISLAVE).EQ.0) CYCLE
         IF(ISTATS(ISLAVE).EQ.150) THEN
           CALL GETSECS(NOW,0)
           IF(NOW-ISTRTME(ISLAVE).GT.30)THEN   ! 30 seconds
             ISTATS(ISLAVE)=0
           END IF
         END IF
150    CONTINUE

! -- Look for the fastest free slave.

       ITEMP2=0
       DO 200 ISLAVE=1,NSLAVE
         IF(IDET(ISLAVE).EQ.0) GO TO 200
         IF(ISTATS(ISLAVE).EQ.0) THEN
           ITEMP2=ITEMP2+1
           IF(ITEMP2.EQ.1) THEN
             ITEMP1=IRUNTME(ISLAVE)
             JJSLAVE=ISLAVE
           ELSE
             IF(IRUNTME(ISLAVE).LT.ITEMP1)THEN
               ITEMP1=IRUNTME(ISLAVE)
               JJSLAVE=ISLAVE
             END IF
           END IF
         END IF
200    CONTINUE
       IF(ITEMP2.NE.0)THEN
         ISLAVE=JJSLAVE

! -- First we make sure that the observation flag file has been deleted.

         AFILE=TRIM(ASLDIR(ISLAVE))//OREADFLE
#ifdef LF90
         OPEN(UNIT=TEMPUNIT2,FILE=AFILE,                    &
         ACTION='READWRITE,DENYNONE',STATUS='OLD',          &
         BLOCKSIZE=1,IOSTAT=IERR)
#else
         OPEN(UNIT=TEMPUNIT2,FILE=AFILE,STATUS='OLD',IOSTAT=IERR)
#endif
         IF(IERR.EQ.0)THEN
           CALL CLOSEFILE(JFAIL,TEMPUNIT2,2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           CALL WAIT(IIWAIT)
           CALL DELFILE(JFAIL,TEMPUNIT2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           GO TO 9300
         END IF
         GO TO 1000
       END IF

C -- See if any runs are complete.

210    DO 500 ISLAVE=1,NSLAVE
         IF(IDET(ISLAVE).EQ.0) GO TO 500
         AFILE=TRIM(ASLDIR(ISLAVE))//OREADFLE
#ifdef LF90
         OPEN(UNIT=TEMPUNIT2,FILE=AFILE,                &
         ACTION='READWRITE,DENYNONE',STATUS='OLD',      &
         BLOCKSIZE=1,IOSTAT=IERR)                       &
#else
         OPEN(UNIT=TEMPUNIT2,FILE=AFILE(1:NBLNK(AFILE)),  &
         STATUS='OLD',IOSTAT=IERR)
#endif
         IF(IERR.EQ.0)THEN
           CALL CLOSEFILE(JFAIL,TEMPUNIT2,2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           CALL WAIT(IIWAIT)
           CALL DELFILE(JFAIL,TEMPUNIT2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           CALL GETSECS(NOW,0)
           IF(ISTATS(ISLAVE).EQ.150) GO TO 500	
           IF(ISTATS(ISLAVE).EQ.-999) THEN !doing a run left over from before
             WRITE(RMRUNIT,220,ERR=9900) RESULT,TRIM(ASLAVE(ISLAVE))
220          FORMAT(3X,A,                                                 &
             ':- slave "',A,'" finished execution; old run so results ',  &
             'not needed.')
             GO TO 250
           END IF
           IF(ISTATR(JRUN(ISLAVE)).EQ.-99) THEN  !already downloaded
             WRITE(RMRUNIT,230,ERR=9900) RESULT,TRIM(ASLAVE(ISLAVE))
230          FORMAT(3X,A,                                                 &
             ':- slave "',A,'" finished execution; same run already ',    &
             'completed by another slave.')
             GO TO 250
           END IF
           WRITE(RMRUNIT,240,ERR=9900) RESULT,TRIM(ASLAVE(ISLAVE))
240        FORMAT(3X,A,                                                   &
           ':- slave "',A,'" finished execution; reading results.')
           NCALL=NCALL+1
           FINAL=0
           ICOUNT=0
245        IITRIAL=ITRIAL(JRUN(ISLAVE))
           CALL PLL_READ_OUT_FILES(IFAIL,ISLAVE,GLOBALUNIT,NCATS,NVEXT,  &
                     EXTNAM,MCVCAT,MCVUSE,EXTVAL,INSTRUCTION)
           IF(IFAIL.NE.0)THEN
             ICOUNT=ICOUNT+1
             IF(ICOUNT.LT.3)THEN
               CALL GETSECS(NOW,0)
               WRITE(RMRUNIT,247,ERR=9900) RESULT,TRIM(ASLAVE(ISLAVE))
247            FORMAT(3X,A,
               ':- problem (either model or communications) ',              &
               'in reading results from slave "',A,'".')                    &
               DO II=1,6
                 CALL WAIT(20)
               END DO
               GO TO 245
             ELSE
               ITRIAL(JRUN(ISLAVE))=ITRIAL(JRUN(ISLAVE))+1
               IF(ITRIAL(JRUN(ISLAVE)).GT.MAXTRIAL) GO TO 9891
               ISTATS(ISLAVE)=0
               ISTATR(JRUN(ISLAVE))=0
               CALL GETSECS(NOW,0)
               WRITE(6,251)TRIM(ASLAVE(ISLAVE))
251            FORMAT('  Problem encountered in reading results of ',      &
               'run on slave "',                                           &
               A,'".',/,'  Will carry out run again.',/)
               WRITE(RMRUNIT,249,ERR=9900) RESULT,TRIM(ASLAVE(ISLAVE))
249            FORMAT(3X,A,':- problems continuing with run on slave "',   &
               A,'". Will run model again.')
               GO TO 50
             END IF
           END IF
           ISTATR(JRUN(ISLAVE))=-99
           IIRUN=IIRUN+1
           ITEMPS=JRUN(ISLAVE)
           IF(NRUN.GT.1) THEN
             CALL WRITINT(ATEMP,IIRUN)
#ifdef LF90
             WRITE(6,255) TRIM(ATEMP)
255          FORMAT('+   ',A,' runs completed.')
#endif
#ifdef LF95
             WRITE(6,255) TRIM(ATEMP)
255          FORMAT('+   ',A,' runs completed.')
#endif
#ifdef UNIX
             WRITE(6,255) TRIM(ATEMP)
255          FORMAT('    ',A,' runs completed.')
#endif
           END IF
250        CONTINUE
           ITEMP1=IRUNTME(ISLAVE)
           IRUNTME(ISLAVE)=NOW-ISTRTME(ISLAVE)
           IF(IRUNTME(ISLAVE).LE.0) IRUNTME(ISLAVE)=1	
           ISTATS(ISLAVE)=0

! -- If this is the first parallel call, all estimated run times are upgraded.

260        CONTINUE
           NNRUN=NNRUN+1
           IF(NNRUN.EQ.1)THEN
             DO 600 JSLAVE=1,NSLAVE
               IF(ISLAVE.NE.JSLAVE)THEN
                 IRUNTME(JSLAVE)=IRUNTME(JSLAVE)*                   &
                 FLOAT(IRUNTME(ISLAVE))/FLOAT(ITEMP1)
                 IF(IRUNTME(JSLAVE).LE.0) IRUNTME(JSLAVE)=1
               END IF
600          CONTINUE
           END IF
           DO 300 IRUN=1,NRUN
             IF(ISTATR(IRUN).GT.-99) GO TO 500  !used to be 50
300        CONTINUE
           GO TO 9000
         END IF
500    CONTINUE

       CALL WAIT(IIWAIT)
       GO TO 50

! -- If a slave is free and there is a run to do we give the slave that run.

1000   CONTINUE

       DO 1100 IRUN=1,NRUN
         IF(ISTATR(IRUN).EQ.0) THEN
           CALL GETSECS(ISTRTME(ISLAVE),0)
           WRITE(RMRUNIT,1070) RESULT,TRIM(ASLAVE(ISLAVE))
1070       FORMAT(3x,A,':- slave "',A,'" commencing model run.')
#ifdef LF90
           CALL FLUSH(RMRUNIT)
#endif
#ifdef LF95
           CALL FLUSH(RMRUNIT)
#endif
#ifdef UNIX
           CALL FLUSH(RMRUNIT)
#endif
           CALL PLL_WRITE_IN_FILES(JFAIL,ISLAVE,NPE,APAR,NPT,IPTR,  &
           PARVALSETS(IRUN,1),PVAL,PRECIS,NOPNT)
           IF(JFAIL.EQ.2) GO TO 9200
           IF(JFAIL.NE.0) GO TO 9891
           CALL WAIT(IIWAIT)
           AFILE=TRIM(ASLDIR(ISLAVE))//PREADFLE
#ifdef LF90
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,                      &
           ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9200)
#else
           OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9200)
#endif
           WRITE(TEMPUNIT,1050,ERR=9199)
1050       FORMAT('P')
           CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           ISTATR(IRUN)=99
           ISTATS(ISLAVE)=99
           JRUN(ISLAVE)=IRUN
           GO TO 50
         END IF
1100   CONTINUE

! -- There is a slave free. Can we redo a run on a faster machine?

       CALL GETSECS(NOW,0)
       ITEMP3=0
       JJSLAVE=0
       DO 1200 JSLAVE=1,NSLAVE
         IF(IDET(JSLAVE).EQ.0) GO TO 1200
         IF(JSLAVE.EQ.ISLAVE) GO TO 1200
         IF(ISTATS(JSLAVE).NE.99) GO TO 1200
         ITEMP4=ISTRTME(JSLAVE)+IRUNTME(JSLAVE)
         IF(ITEMP4.GT.ITEMP3)THEN
           ITEMP3=ITEMP4
           JJSLAVE=JSLAVE
         END IF
1200   CONTINUE
       IF(JJSLAVE.EQ.0) GO TO 1500
       ITEMP3=ITEMP3-NOW
       IF(ITEMP3.LT.2) GO TO 1500		! too small to worry about
       IF(FLOAT(ITEMP3).GT.1.5*FLOAT(IRUNTME(ISLAVE)))THEN
         IRUN=JRUN(JJSLAVE)
         ISTRTME(ISLAVE)=NOW
         WRITE(RMRUNIT,1230) RESULT,                 &
         TRIM(ASLAVE(ISLAVE)),TRIM(ASLAVE(JJSLAVE))
1230     FORMAT(3X,A,                                             &
         ':- slave "',A,'" commencing model run because it can ', &
         'complete it before slave "',A,'".')
         CALL PLL_WRITE_IN_FILES(JFAIL,ISLAVE,NPE,APAR,NPT,IPTR,  &
         PARVALSETS(IRUN,1),PVAL,PRECIS,NOPNT)
         IF(JFAIL.EQ.2) GO TO 9200
         IF(JFAIL.NE.0) GO TO 9891
         CALL WAIT(IIWAIT)
         AFILE=TRIM(ASLDIR(ISLAVE))//PREADFLE
#ifdef LF90
         OPEN(UNIT=TEMPUNIT,FILE=AFILE,                       &
         ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9200)
#else
         OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9200)
#endif
         WRITE(TEMPUNIT,1050,ERR=9199)
         CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
         IF(JFAIL.NE.0) GO TO 9891
         ISTATR(IRUN)=99
         ISTATS(ISLAVE)=99
         JRUN(ISLAVE)=IRUN
         ISTATS(JJSLAVE)=100			!so only one slave does it
         GO TO 210
       END IF

! -- Are any runs overdue?

1500   CONTINUE
       ITEMP3=0
       JJSLAVE=0
       DO 1700 JSLAVE=1,NSLAVE
         IF(IDET(JSLAVE).EQ.0) GO TO 1700
         IF(JSLAVE.EQ.ISLAVE) GO TO 1700
         IF(ISTATS(JSLAVE).NE.99) GO TO 1700
         ITEMP4=ISTRTME(JSLAVE)+IRUNTME(JSLAVE)-NOW
         IF(ITEMP4.LT.ITEMP3) THEN
           ITEMP3=ITEMP4
           JJSLAVE=JSLAVE
         END IF
1700   CONTINUE
       IF((JJSLAVE.EQ.0).OR.                                  &
          (FLOAT(ITEMP3).GT.-0.5*FLOAT(IRUNTME(ISLAVE)))) THEN	!arbitrary
          GO TO 210
       END IF
       IRUN=JRUN(JJSLAVE)
       ISTRTME(ISLAVE)=NOW
       WRITE(RMRUNIT,1520) RESULT,TRIM(ASLAVE(ISLAVE)),       &
       TRIM(ASLAVE(JJSLAVE))
1520   FORMAT(3X,A,                                              &
       ':- slave "',a,'" commencing model run because slave "',  &
       A,'" is overdue.')
       CALL PLL_WRITE_IN_FILES(JFAIL,ISLAVE,NPE,APAR,NPT,IPTR,  &
       PARVALSETS(IRUN,1),PVAL,PRECIS,NOPNT)
       IF(JFAIL.EQ.2) GO TO 9200
       IF(JFAIL.NE.0) GO TO 9891
       CALL WAIT(IIWAIT)
       AFILE=TRIM(ASLDIR(ISLAVE))//PREADFLE
#ifdef LF90
       OPEN(UNIT=TEMPUNIT,FILE=AFILE,                     &
       ACTION='READWRITE,DENYNONE',BLOCKSIZE=1,ERR=9200)
#else
       OPEN(UNIT=TEMPUNIT,FILE=AFILE,ERR=9200)
#endif
       WRITE(TEMPUNIT,1050,ERR=9199)
       CALL CLOSEFILE(JFAIL,TEMPUNIT,2,AFILE)
       IF(JFAIL.NE.0) GO TO 9891
       ISTATR(IRUN)=99
       ISTATS(ISLAVE)=99
       JRUN(ISLAVE)=IRUN
       ISTATS(JJSLAVE)=100		!so only one slave does it
       GO TO 210

! -- Before returning we re-assign the status of incompleted runs.
! -- Also, the signal files are deleted. If they re-appear PEST will
!    use these slaves again.

9000   DO 9100 ISLAVE=1,NSLAVE
         IF(IDET(ISLAVE).EQ.0) GO TO 9100
         IF((ISTATS(ISLAVE).EQ.99).OR.(ISTATS(ISLAVE).EQ.100))THEN
           AFILE=TRIM(ASLDIR(ISLAVE))//SREADFLE
           CALL DELFILE(JFAIL,-TEMPUNIT2,AFILE)
           IF(JFAIL.NE.0) GO TO 9891
           ISTATS(ISLAVE)=-999
         END IF
9100   CONTINUE
       GO TO 9999

9199   CALL CLOSEFILE(JFAIL,TEMPUNIT,0,AFILE)
       IF(JFAIL.NE.0) GO TO 9891
9200   CALL GETSECS(ISTRTME(ISLAVE),0)
       WRITE(RMRUNIT,9210) RESULT,TRIM(ASLAVE(ISLAVE))
9210   FORMAT(3X,A,                                                &
       ':- cannot write to slave "',A,'" to commence model run.')
       ISTATS(ISLAVE)=150
       GO TO 50

9300   CALL GETSECS(ISTRTME(ISLAVE),0)
       WRITE(RMRUNIT,9310) RESULT,TRIM(ASLAVE(ISLAVE))
9310   FORMAT(3X,A,                                                &
       ':- previous observ.rdy file on slave "',A,'" not deleted.')
       ISTATS(ISLAVE)=150
       GO TO 50

9900   CONTINUE
       CALL ADDQUOTE(RMRFILE,AFILE)
       WRITE(AMESSAGE,9910) TRIM(AFILE)
9910   FORMAT('Cannot write to run management file ',A,'.')
       GO TO 9891
	
9400   CONTINUE
       WRITE(AMESSAGE,9410) TRIM(AFILE)
9410   FORMAT('Cannot communicate with message file ',A,'.')
       GO TO 9891
	
9999   CONTINUE
       RETURN

9891   IFAIL=1
       RETURN

       END SUBROUTINE PLL_DO_RUNS


       SUBROUTINE PLL_FIN(IFAIL)

! -- Subroutine PLL_FIN deallocates memory and closes files.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       IMPLICIT NONE

       INTEGER, INTENT(OUT)    :: IFAIL     ! Failure flag

! Initialisation

       IFAIL=0

! -- The run management record file is closed.

       CLOSE(UNIT=RMRUNIT,IOSTAT=IERR)
       	
! -- Allocated arrays are de-allocated.

       IF(ALLOCATED(IDET))DEALLOCATE(IDET,STAT=IERR)
       IF(ALLOCATED(ISTATS))DEALLOCATE(ISTATS,STAT=IERR)
       IF(ALLOCATED(ISTATR))DEALLOCATE(ISTATR,STAT=IERR)
       IF(ALLOCATED(IRUNTME))DEALLOCATE(IRUNTME,STAT=IERR)
       IF(ALLOCATED(ISTRTME))DEALLOCATE(ISTRTME,STAT=IERR)
       IF(ALLOCATED(JRUN))DEALLOCATE(JRUN,STAT=IERR)
       IF(ALLOCATED(ITRIAL))DEALLOCATE(ITRIAL,STAT=IERR)
       IF(ALLOCATED(ASLAVE))DEALLOCATE(ASLAVE,STAT=IERR)
       IF(ALLOCATED(ASLDIR))DEALLOCATE(ASLDIR,STAT=IERR)
       IF(ALLOCATED(RUNPARVAL))DEALLOCATE(RUNPARVAL,STAT=IERR)
       IF(ALLOCATED(SCOM))DEALLOCATE(SCOM,STAT=IERR)

       RETURN

       END SUBROUTINE PLL_FIN




       SUBROUTINE PLL_WRITE_IN_FILES(IFAIL,ISLAVE,NPE,APAR,NPT,IPTR,PARVALSET,PVAL,  &
       PRECIS,NOPNT)

! -- Subroutine PLL_WRITE_IN_FILES writes model input files on a slave.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       USE MODEL_IO
       IMPLICIT NONE

       INTEGER, INTENT(OUT)    :: IFAIL     ! Failure flag
       INTEGER, INTENT(IN)     :: ISLAVE    ! Name of slave to do runs on
       INTEGER, INTENT(IN)     :: NPE       ! No. of adjustable parameters
       INTEGER, INTENT(IN)     :: NPT       ! Total no. of parameters
       INTEGER, INTENT(IN)     :: PRECIS    ! Precision protocol
       INTEGER, INTENT(OUT)    :: NOPNT     ! Decimal point protocol
       INTEGER, INTENT(IN)     :: IPTR(NPE) ! Pointer in PVAL to each adj. param.
       DOUBLE PRECISION, INTENT(IN)  :: PARVALSET(NPE) ! Adj. params for this run
       CHARACTER (LEN=*), INTENT(IN) :: APAR(NPT)   ! Parameter names
       DOUBLE PRECISION, INTENT(IN)  :: PVAL(NPT)   ! Current unperterbed param values

       INTEGER                 :: IFAIL,JFAIL

! -- Initialisation

       IFAIL=0

! -- First an array is filled of parameter values for the current run.

       DO I=1,NPT
         RUNPARVAL(I)=PVAL(I)
       END DO
       DO I=1,NPE
         RUNPARVAL(I)=PARVALSET(IPTR(I))
       END DO

!NOTE - I have to initialise MIO somewhere - in main program

! -- Next the names of model input files are changed.

       DO I=1,NUMINFILE
         CALL PLL_FORM_FILENAME(JFAIL,ASLDIR(ISLAVE),MODEL_INPUT_FILE(I) &
         AFILE)
         IF(JFAIL.NE.0) GO TO 9890
         CALL MIO_ADA_MODINFILE_CHANGE(JFAIL,I,AFILE)
       END DO

! -- Model input files are written.

       CALL MIO_ADA_WRITEFILES(JFAIL,NPT,AFILE,NOPNT,NW,PRECIS,RUNPARVAL)
       IF(JFAIL.NE.0)THEN
         IF(INDEX(AMESSAGE,'input file').NE.0)THEN
           IFAIL=2
           RETURN
         ELSE
           GO TO 9890
         RETURN
       END IF

       RETURN

9890   IFAIL=1
       RETURN

       END SUBROUTINE PLL_WRITE_IN_FILES





       SUBROUTINE PLL_READ_OUT_FILES(IFAIL,ISLAVE,GLOBALUNIT,NCATS,NVEXT,  &
                     EXTNAM,MCVCAT,MCVUSE,EXTVAL,INSTRUCTION)

! -- Subroutine PLL_READ_OUT_FILES reads model output files from a slave.

       USE GLOBAL_DATA, ONLY: AMESSAGE,ERRSUB
       USE MODEL_IO
       IMPLICIT NONE

       INTEGER, INTENT(OUT)          :: IFAIL     ! Failure flag
       INTEGER, INTENT(IN)           :: ISLAVE    ! Name of slave to do runs on
       INTEGER, INTENT(IN)           :: NCATS  ! Number of categories of model-calculated values
       INTEGER, INTENT(IN)           :: NVEXT  ! number of values to extract
       CHARACTER (LEN=*), INTENT(IN) :: EXTNAM(NVEXT) ! extracted-value names
       CHARACTER (LEN=6), INTENT(IN) :: MCVCAT(NCATS) ! Categories supported by application
       LOGICAL, INTENT(IN)           :: MCVUSE(NCATS) ! Do extractions for this category?
       DOUBLE PRECISION, INTENT(OUT) :: EXTVAL(NVEXT) ! extracted values
       CHARACTER (LEN=*),INTENT(OUT) :: INSTRUCTION ! instruction


       INTEGER     :: JFAIL,I


! -- Initialisation

       IFAIL=0

! -- The names of model outpuT files are changed.

       DO I=1,NUMOUTFILE
         CALL PLL_FORM_FILENAME(JFAIL,ASLDIR(ISLAVE),MODEL_OUTPUT_FILE(I) &
         AFILE)
         IF(JFAIL.NE.0) GO TO 9890
         CALL MIO_EXT_MODOUTFILE_CHANGE(JFAIL,I,AFILE)
         IF(JFAIL.NE.0) GO TO 9890
       END DO

! -- Model output files are read.

       CALL MIO_EXT(JFAIL,GLOBALUNIT,NCATS,NVEXT,EXTNAM,MCVCAT,MCVUSE,EXTVAL,   &
                     INSTRUCTION)

       IF(JFAIL.NE.0)IFAIL=1
       RETURN

9890   IFAIL=1
       RETURN

       END SUBROUTINE PLL_READ_OUT_FILES




        SUBROUTINE WRITINT(ATEMP,IVAL)

!       Subroutine WRITINT writes an integer to a character variable.

        INTEGER, INTENT(IN)    :: IVAL  ! The integer to be written.
        CHARACTER (LEN=*)      :: ATEMP ! The character variable to write it to

        CHARACTER (LEN=6)      :: AFMT

	AFMT='(I   )'
	WRITE(AFMT(3:5),'(I3)') LEN(ATEMP)
	WRITE(ATEMP,AFMT)IVAL
        ATEMP=ADJUSTL(ATEMP)
	RETURN
	END SUBROUTINE WRITINT



       SUBROUTINE DELFILE(IFAIL,IUNIT,FILENAME)

       USE GLOBAL_DATA, ONLY: AMESSAGE
       IMPLICIT NONE

       INTEGER, INTENT(OUT)     :: IFAIL     ! Failure index
       INTEGER, INTENT(IN)      :: IUNIT     ! Unit number of open file
       CHARACTER (LEN=*)        :: FILENAME  ! Name of file to be deleted
	
	
       LOGICAL     :: LEXIST
       INTEGER     :: ICOUNT,JUNIT,IERR,IIWAIT,JFAIL
#ifdef UNIX
       INTEGER     :: UNLINK,ITEMP
#endif	

       IFAIL=0
       IIWAIT=IWAIT
       JUNIT=IUNIT
       IF(IUNIT.LT.0)THEN
         INQUIRE(FILE=FILENAME(1:NBLNK(FILENAME)),EXIST=LEXIST)
         IF(.NOT.LEXIST) RETURN
         JUNIT=-JUNIT
       END IF
       ICOUNT=0
#ifdef UNIX
10     ITEMP=UNLINK(FILENAME(1:NBLNK(FILENAME)))
#else
10     CALL SYSTEM('DEL "'//TRIM(FILENAME)//'" > NUL')
#endif
       CALL WAIT(IIWAIT)
#ifdef LF90
       OPEN(UNIT=JUNIT,FILE=FILENAME(1:NBLNK(FILENAME)),  &
       ACTION='READWRITE,DENYNONE',STATUS='OLD',          &
       BLOCKSIZE=1,IOSTAT=IERR)
#else
       OPEN(UNIT=JUNIT,FILE=FILENAME(1:NBLNK(FILENAME)),  &
       STATUS='OLD',IOSTAT=IERR)
#endif
       IF(IERR.NE.0) RETURN
       CALL WAIT(IIWAIT)
       CALL CLOSEFILE(JFAIL,JUNIT,2,FILENAME)
       IF(JFAIL.NE.0) GO TO 9891
       ICOUNT=ICOUNT+1
       IF(ICOUNT.GT.10)THEN
         CALL ADDQUOTE(FILENAME,AFILE)
         WRITE(AMESSAGE,20) TRIM(AFILE)
20       FORMAT('Error in parallel communications: cannot delete file ',A)
         GO TO 9891
       END IF
       CALL WAIT(IIWAIT)
       GO TO 10

9891   IFAIL=1
       RETURN

       END SUBROUTINE DELFILE


       SUBROUTINE CLOSEFILE(IFAIL,IUNIT,ISTATUS,FILENAME)

       USE GLOBAL_DATA, ONLY: AMESSAGE
       IMPLICIT NONE

       INTEGER, INTENT(OUT)          :: IFAIL     ! Failure index
       INTEGER, INTENT(IN)           :: IUNIT     ! Unit number of open file
       INTEGER, INTENT(IN)           :: ISTATUS   ! Set to 0 if not sure if opened.
       CHARACTER (LEN=*), INTENT(IN) :: FILENAME  ! Name of file to close


       LOGICAL    :: LOPENED
       INTEGER    :: ICOUNT,IERR,ISTATUS,IIWAIT

       IFAIL=0
       IIWAIT=IWAIT
       IF(ISTATUS.EQ.2) GO TO 90
       INQUIRE(UNIT=IUNIT,OPENED=LOPENED)
       IF(.NOT.LOPENED) RETURN
90     ICOUNT=0
100    ICOUNT=ICOUNT+1
       CLOSE(UNIT=IUNIT,IOSTAT=IERR)
       IF(IERR.NE.0) THEN	
         IF(ICOUNT.GT.20) THEN
           CALL ADDQUOTE(FILENAME,AFILE)
           WRITE(AMESSAGE,110)
110        FORMAT('Error in parallel communications: cannot close ', &
           'file ',A)
           GO TO 9891
         END IF
         CALL WAIT(IIWAIT)
         GO TO 100
       END IF

       RETURN

9891   IFAIL=1
       RETURN
       END SUBROUTINE CLOSEFILE



       SUBROUTINE WRITE_SLAVE_DATA_1(IFAIL)

       USE GLOBAL_DATA, ONLY: AMESSAGE
       IMPLICIT NONE

       INTEGER, INTENT(OUT)   :: IFAIL
       INTEGER                :: ISLAVE

       IFAIL=0

       WRITE(RMRUNIT,20,ERR=9900)
20     FORMAT(/,' SLAVE DETAILS:-',/)
       WRITE(RMRUNIT,50,ERR=9900)
50     FORMAT(' Slave Name',T30,'Slave Working Directory')
       WRITE(RMRUNIT,55,ERR=9900)
55     FORMAT(' ----------',T30,'------------------------')
       DO ISLAVE=1,NSLAVE
         WRITE(RMRUNIT,60,ERR=9900) TRIM(ASLAVE(ISLAVE)),   &
         TRIM(ASLDIR(ISLAVE))
60       FORMAT(' "',A,'"',T30,A)
       END DO
       WRITE(RMRUNIT,110,ERR=9900)
110    FORMAT(/,/,' Attempting to communicate with slaves ....',/)

       RETURN

9900   CONTINUE
       CALL ADDQUOTE(RMRFILE,AFILE)
       WRITE(AMESSAGE,9910) TRIM(AFILE)
9910   FORMAT('Cannot write to file ',A)
       IFAIL=1
       RETURN

       END SUBROUTINE WRITE_SLAVE_DATA_1




       SUBROUTINE ADDQUOTE(AFILE,AQFILE)

C -- Subroutine ADDQUOTE adds quotes to a filename if it has a space in it.

       CHARACTER (LEN=*), INTENT(IN)   :: AFILE
       CHARACTER (LEN=*), INTENT(OUT)  :: AQFILE
       IMPLICT NONE
       INTEGER           :: NBB

       IF(INDEX(TRIM(AFILE),' ').EQ.0)THEN
         AQFILE=AFILE
       ELSE
         AQFILE(1:1)='"'
         AQFILE(2:)=TRIM(AFILE)
         NBB=LEN_TRIM(AQFILE)+1
         AQFILE(NBB:NBB)='"'
       END IF

       RETURN
       END SUBROUTINE ADDQUOTE



       SUBROUTINE WAIT(NHUNDSEC)

! -- Subroutine WAIT causes the executing program to wait for NHUNDSEC
!    hundredths of a second.

       IMPLICIT NONE

       INTEGER, INTENT(IN)   :: NHUNDSEC
       INTEGER               :: DDATE(8),ITICKS,ITICKS1

       CALL DATE_AND_TIME(VALUES=DDATE)
       ITICKS=DDATE(5)*360000+DDATE(6)*6000+DDATE(7)*100+DDATE(8)/10
10     CALL DATE_AND_TIME(VALUES=DDATE)
       ITICKS1=DDATE(5)*360000+DDATE(6)*6000+DDATE(7)*100+DDATE(8)/10
       IF(ITICKS1.LT.ITICKS) ITICKS1=ITICKS1+8640000
       IF(ITICKS1.LT.ITICKS+NHUNDSEC) GO TO 10

       RETURN

       END SUBROUTINE WAIT




       SUBROUTINE PLL_FORM_FILENAME(IFAIL,DIRECTORY,INFILE,OUTFILE)

       IMPLICIT NONE
       INTEGER, INTENT(OUT)           :: IFAIL
       CHARACTER (LEN=*), INTENT(IN)  :: DIRECTORY
       CHARACTER (LEN=*), INTENT(IN)  :: INFILE
       CHARACTER (LEN=*), INTENT(OUT) :: OUTFILE

       IFAIL=0
       OUTFILE=TRIM(DIRECTORY)//TRIM(INFILE)

       RETURN
       END PLL_FORM_FILENAME



       SUBROUTINE WRITE_SLAVE_DATA_2(IFAIL)

       USE GLOBAL_DATA, ONLY: AMESSAGE
       IMPLICIT NONE

       INTEGER, INTENT(OUT)   :: IFAIL

       INTEGER             :: JFAIL,I
       CHARACTER (LEN=10)  :: IWAIT


       IFAIL=0
       WRITE(RMRUNIT,100)
100    FORMAT(/,/,' SLAVE MODEL INPUT AND OUTPUT FILES:-')
       DO 300 ISLAVE=1,NSLAVE
         WRITE(RMRUNIT,120) TRIM(ASLAVE(ISLAVE))
120      FORMAT(/,' Slave "',a,'" ----->')
         WRITE(RMRUNIT,130) TRIM(ASLAVE(ISLAVE))
130      FORMAT(/,'     Model input files on slave "',A,'":-')
         DO 170 I=1,NUMINFILE
           CALL PLL_FORM_FILENAME(JFAIL,ASLDIR(ISLAVE),    &
           MODEL_INPUT_FILE(I),SINFLE)
           IF(JFAIL.NE.0) GO TO 9891
           CALL ADDQUOTE(SINFLE,AFILE)
           WRITE(RMRUNIT,150) TRIM(AFILE)
150        FORMAT(T10,A)
170      CONTINUE
         WRITE(RMRUNIT,180) TRIM(ASLAVE(ISLAVE))
180      FORMAT(/,'     Model output files on slave "',A,'":-')
         DO 200 I=1,NUMOUTFILE
           CALL PLL_FORM_FILENAME(JFAIL,ASLDIR(ISLAVE),    &
           MODEL_OUTFILE_FILE(I),SOUFLE)
           IF(JFAI.NE.0) GO TO 9891
           CALL ADDQUOTE(SOUFLE,AFILE)
           WRITE(RMRUNIT,150) TRIM(AFILE)
           CALL DELFILE(JFAIL,-TEMPUNIT,SOUFLE)
           IF(JFAIL.NE.0) GO TO 9891
200      CONTINUE
         WRITE(RMRUNIT,250) TRIM(ASLAVE(ISLAVE))
250      FORMAT(/,'     Model command line for slave "',A,'":-')
         CALL ADDQUOTE(SCOM(ISLAVE),AFILE)
         WRITE(RMRUNIT,150) TRIM(AFILE)
         WRITE(RMRUNIT,*)
300    CONTINUE

       CALL WRITINT(ATEMP,IWAIT)
       WRITE(RMRUNIT,310) TRIM(ATEMP)
310    FORMAT(/,' Average wait interval    : ',A,   &
       ' hundredths of a second.')

       WRITE(RMRUNIT,350)
350    FORMAT(/,/,T29,'RUN MANAGEMENT RECORD')

       RETURN

9891   IFAIL=1
       RETURN	
       END SUBROUTINE WRITE_SLAVE_DATA_2


       SUBROUTINE GETSECS(ISECS,INIT)

! -- Subroutine GETSETS calculates elapsed seconds from an initial time
!    (expressed in seconds) to a the current moment.

       IMPLICIT NONE
       INTEGER, INTENT(OUT)        :: ISECS
       INTEGER, INTENT(IN)         :: INIT

       INTEGER            :: MM,DD,YY,HH,MIN,NDAYS
       REAL               :: SS
       CHARACTER (LEN=11) :: RESULT1,RESULT2

       CALL DATE_AND_TIME(DATE=RESULT1,TIME=RESULT2)
       READ(RESULT1,'(I4,I2,I2)') YY,MM,DD
       READ(RESULT2,'(I2,I2,F6.0)') HH,MIN,SS
       WRITE(RESULT,'(I2.2,'':'',I2.2,'':'',F5.2)') HH,MIN,SS
       IF(INIT.EQ.0) THEN
         NDAYS=NUMDAYS(D1,M1,Y1,DD,MM,YY)
         ISECS=NDAYS*86400+(HH-H1)*3600+(MIN-MIN1)*60+SS-SS1
       ELSE
         M1=MM
         D1=DD
         Y1=YY
         H1=HH
         MIN1=MIN
         SS1=SS
       END IF

       RETURN
       END SUBROUTINE GETSECS



       INTEGER FUNCTION NUMDAYS(DR,MR,YR,D,M,Y)

       IMPLICIT NONE

! -- Function numdays calculates the number of days between dates
!    D-M-Y and DR-MR-YR. If the former preceeds the latter the answer is
!    negative.

! -- Arguments are as follows:-
!       dr,mr,yr:     days, months and years of first date
!       d,m,y:        days, months and years of second date
!       numdays returns the number of elapsed days

       INTEGER, INTENT(IN)   :: DR,MR,YR
       INTEGER, INTENT(OUT)  :: D,M,Y

       INTEGER       :: FLAG,I,J,DA(12),YE,ME,DE,YL,ML,DL

       DATA DA /31,28,31,30,31,30,31,31,30,31,30,31/

! -- The smaller of the two dates is now chosen to do the counting from.

       IF(Y.LT.YR)GO TO 10
       IF((Y.EQ.YR).AND.(M.LT.MR)) GO TO 10
       IF((Y.EQ.YR).AND.(M.EQ.MR).AND.(D.LT.DR)) GO TO 10
       FLAG=0
       YE=YR
       ME=MR
       DE=DR
       YL=Y
       ML=M
       DL=D
       GO TO 20
10     FLAG=1
       YE=Y
       ME=M
       DE=D
       YL=YR
       ML=MR
       DL=DR

! -- In the above the postscript "E" stands for earlier date, while
!    "L" stands for the later date.

20     NUMDAYS=0
       IF((ME.EQ.ML).AND.(YL.EQ.YE))THEN
         NUMDAYS=DL-DE
         IF(FLAG.EQ.1) NUMDAYS=-NUMDAYS
         RETURN
       END IF

       DO 30 J=ME,12
         IF((ML.EQ.J).AND.(YE.EQ.YL))GOTO 40
         NUMDAYS=NUMDAYS+DA(J)
         IF((J.EQ.2).AND.(LEAP(YE)))NUMDAYS=NUMDAYS+1
30     CONTINUE
       GO TO 50
40     NUMDAYS=NUMDAYS+DL-DE
       IF(FLAG.EQ.1)NUMDAYS=-NUMDAYS
       RETURN

50     DO 60 I=YE+1,YL
         DO 70 J=1,12
           IF((YL.EQ.I).AND.(ML.EQ.J))GO TO 80
           NUMDAYS=NUMDAYS+DA(J)
           IF((J.EQ.2).AND.(LEAP(I))) NUMDAYS=NUMDAYS+1
70       CONTINUE
60     CONTINUE
       WRITE(6,65)
65     FORMAT(/,' ERROR IN SUBROUTINE NUMDAYS')
       STOP

80     NUMDAYS=NUMDAYS+DL-DE
       IF(FLAG.EQ.1) NUMDAYS=-NUMDAYS

       RETURN
       END FUNCTION NUMDAYS




       LOGICAL FUNCTION LEAP(YEAR)

C -- Function LEAP returns .true. if a year is a leap year.

       IMPLICIT NONE
       INTEGER YEAR

       LEAP = ( MOD(YEAR,4).EQ.0 .AND. MOD(YEAR,100).NE.0 ) .OR.  &
              ( MOD(YEAR,400).EQ.0 .AND. YEAR.NE.0 )

       RETURN
       END FUNCTION LEAP






! STUFF TO DO
!1. AT THE MOMENT I USE LAHEY PROTOCOL FOR OPENING. PUT ALL OPEN STATEMENTS
!     INTO A SINGLE SUBROUTINE TO MAKE THIS PROTOCOL EASIER TO CHANGE. GO
!     OVER THE ORIGINAL MODULE TO ENSURE THAT I'VE MAKE ALLOWANCE FOR ALL
!     UNIX STUFF.
!2. LOOK FOR 79 AND REPLACE WITH TEMPUNIT
!3. LOOK FOR 80 AND REPLACE WITH RMRUNIT
!4. LOOK FOR NBLNK AND GET RID OF IT
!5. MAKE SUBROUTINE PLL_FORM_FILENAME MORE EFFICIENT. ALLOW IT TO HANDLE
!     CASES WHERE WE HAVE ../ OR ..\ IN MODEL INPUT AND OUTPUT FILENAMES
!6. DO NOT ALLOW DO_RUNS TO BE CALLED BEFORE SUBROUTINE PLL_INI_COMM CALLED.
!7. MAKE SURE THAT WE HAVE THE ABILITY TO MAKE A FEW ATTEMPTS TO READ MODEL
!     OUTPUT FILES.
!8. we may have to put in changes to the mio module to make sure that files
!     are properly closed - perhaps put an alternative closing subroutine
!     in the utility module (ie. closefile), and use it if we are running in
!     parallel mode. We'll need a utility flag for this.
!9. At present, after a call to the model output reading model, we try a
!     few times if there is a failure. Maybe we should check the INSTRUCTION
!     string. If a bad instruction is the reason (and hence this is not empty)
!     then we should just stop execution. (Make sure that this there will
!     not be occasions however when the INSTRUCTION string still has contents
!     but that we should stop anyway.)
